/*
 * search.c  -- isearch
 *
 * $Id: search.c,v 1.10 2004/05/20 10:25:03 hos Exp $
 *
 */


#include "main.h"
#include "regexp.h"
#include "util.h"
#include <errno.h>


#define WM_APP_FWD (WM_APP + 1)
#define WM_APP_BWD (WM_APP + 2)
#define WM_APP_CLOSE (WM_APP + 3)
#define WM_APP_SEL_CLICK (WM_APP + 4)


typedef struct wnd_ctx {
    IHTMLDocument2 *doc;
    BSTR word;
    int dir;
    long flags;

    BOOL failing;
    BOOL wrapped;

    HWND main;
    HWND stat;
    HWND edit;

    migemo *mp;
    IRegExp *regexp;
} wnd_ctx_t;

typedef LRESULT (*msg_proc_t)(HWND, UINT, WPARAM, LPARAM, wnd_ctx_t *);


static TCHAR *window_class_name = _T("ie-migemo-search");
static TCHAR *window_title_name = _T("ie-migemo-search");
static int margin = 4;


static HRESULT get_active_document(IHTMLDocument2 *doc,
                                   IHTMLDocument2 **active_doc)
{
    IHTMLElement *elem = NULL;
    BSTR str;
    HRESULT hres = S_OK;

    {
        IHTMLDocument2 *inner_doc;

        inner_doc = doc;
        IHTMLDocument2_AddRef(doc);

        while(1) {
            hres = IHTMLDocument2_get_activeElement(inner_doc, &elem);
            if(FAILED(hres)) {
                IHTMLDocument2_Release(inner_doc);
                goto func_end;
            }

            if(elem == NULL) {
                break;
            }

            hres = IHTMLElement_get_tagName(elem, &str);
            if(FAILED(hres)) {
                IHTMLDocument2_Release(inner_doc);
                goto func_end;
            }

            if(str == NULL ||
               (wcsstr(str, L"frame") == NULL &&
                wcsstr(str, L"FRAME") == NULL)) {
                SysFreeString(str);
                break;
            }

            SysFreeString(str);

            {
                IHTMLFrameBase2 *frame;
                IHTMLWindow2 *window;
                IHTMLDocument2 *next_doc;

                hres = IHTMLElement_QueryInterface(elem, &IID_IHTMLFrameBase2,
                                                   (void **)(void *)&frame);
                if(FAILED(hres)) {
                    IHTMLDocument2_Release(inner_doc);
                    goto func_end;
                }

                hres = IHTMLFrameBase2_get_contentWindow(frame, &window);
                IHTMLFrameBase2_Release(frame);
                if(FAILED(hres)) {
                    IHTMLDocument2_Release(inner_doc);
                    goto func_end;
                }

                hres = IHTMLWindow2_get_document(window, &next_doc);
                IHTMLWindow2_Release(window);
                if(FAILED(hres)) {
                    IHTMLDocument2_Release(inner_doc);
                    goto func_end;
                }

                IHTMLElement_Release(elem);
                elem = NULL;

                IHTMLDocument2_Release(inner_doc);
                inner_doc = next_doc;
            }
        }

        *active_doc = inner_doc;
    }

  func_end:
    if(elem != NULL) IHTMLElement_Release(elem);

    return hres;
}

static HRESULT get_selected_range(IHTMLDocument2 *doc, IHTMLTxtRange **range)
{
    IHTMLSelectionObject *sel = NULL;
    IDispatch *range_dp = NULL;
    HRESULT hres = S_OK;

    hres = IHTMLDocument2_get_selection(doc, &sel);
    if(FAILED(hres)) {
        return hres;
    }

    hres = IHTMLSelectionObject_createRange(sel, &range_dp);
    if(FAILED(hres)) {
        goto func_end;
    }

    hres = IDispatch_QueryInterface(range_dp, &IID_IHTMLTxtRange,
                                    (void **)(void *)range);
    if(FAILED(hres)) {
        goto func_end;
    }

  func_end:
    if(sel != NULL) IHTMLSelectionObject_Release(sel);
    if(range_dp != NULL) IDispatch_Release(range_dp);

    return hres;
}

static HRESULT get_selected_text(IHTMLDocument2 *doc, BSTR *str)
{
    IHTMLTxtRange *range;
    HRESULT hres;

    hres = get_selected_range(doc, &range);
    if(FAILED(hres)) {
        return hres;
    }

    hres = IHTMLTxtRange_get_text(range, str);
    IHTMLTxtRange_Release(range);
    if(FAILED(hres)) {
        return hres;
    }

    return S_OK;
}


static unsigned char *mbsdup_from_bstr(BSTR str)
{
    int size;
    unsigned char *mbs;

    size = WideCharToMultiByte(CP_OEMCP, 0,
                               str, wcslen(str), NULL, 0,
                               NULL, NULL);

    mbs = (unsigned char *)malloc(size + 1);
    if(mbs == NULL) {
        return NULL;
    }
    memset(mbs, 0, size + 1);

    WideCharToMultiByte(CP_OEMCP, 0,
                        str, wcslen(str), mbs, size + 1,
                        NULL, NULL);

    return mbs;
}

static BSTR bstrdup_from_mbs(unsigned char *mbs)
{
    int size;
    BSTR str;

    size = MultiByteToWideChar(CP_OEMCP, 0, mbs, strlen(mbs), NULL, 0);

    str = SysAllocStringLen(NULL, size);
    if(str == NULL) {
        return NULL;
    }

    MultiByteToWideChar(CP_OEMCP, 0, mbs, strlen(mbs), str, size + 1);

    return str;
}


static HRESULT find_text_with_migemo(wnd_ctx_t *ctx, IHTMLTxtRange *range,
                                     VARIANT_BOOL *res)
{
    IMatchCollection *matches;
    HRESULT hres;

    {
        unsigned char *word_mb, *pattern_mb;

        word_mb = mbsdup_from_bstr(ctx->word);
        if(word_mb == NULL) {
            return E_FAIL;
        }

        pattern_mb = migemo_query(ctx->mp, word_mb);
        free(word_mb);
        if(pattern_mb != NULL) {
            BSTR pattern;

            pattern = bstrdup_from_mbs(pattern_mb);
            migemo_release(ctx->mp, pattern_mb);
            if(pattern == NULL) {
                return E_FAIL;
            }

            hres = IRegExp_put_Pattern(ctx->regexp, pattern);
            SysFreeString(pattern);
            if(FAILED(hres)) {
                return hres;
            }
        } else {
            hres = IRegExp_put_Pattern(ctx->regexp, ctx->word);
            if(FAILED(hres)) {
                return hres;
            }
        }
    }

    {
        BSTR text;
        IDispatch *matches_dp;

        hres = IHTMLTxtRange_get_text(range, &text);
        if(FAILED(hres)) {
            return hres;
        }

        hres = IRegExp_Execute(ctx->regexp, text, &matches_dp);
        SysFreeString(text);
        if(FAILED(hres)) {
            *res = VARIANT_FALSE;
            return S_OK;
        }

        hres = IDispatch_QueryInterface(matches_dp, &IID_IMatchCollection,
                                        (void **)(void *)&matches);
        IDispatch_Release(matches_dp);
        if(FAILED(hres)) {
            return hres;
        }
    }

    {
        long cnt, i, idx;
        IDispatch *match_dp;
        IMatch *match;
        BSTR match_str;
        VARIANT_BOOL find_res;

        hres = IMatchCollection_get_Count(matches, &cnt);
        if(FAILED(hres)) {
            IMatchCollection_Release(matches);
            return hres;
        }

        for(i = 0; i < cnt; i++) {
            idx = (ctx->dir > 0 ? i : cnt - i - 1);

            hres = IMatchCollection_get_Item(matches, idx, &match_dp);
            if(FAILED(hres)) {
                IMatchCollection_Release(matches);
                return hres;
            }

            hres = IMatchCollection_QueryInterface(match_dp, &IID_IMatch,
                                                   (void **)(void *)&match);
            IDispatch_Release(match_dp);
            if(FAILED(hres)) {
                IMatchCollection_Release(matches);
                return hres;
            }

            hres = IMatch_get_Value(match, &match_str);
            IMatch_Release(match);
            if(FAILED(hres)) {
                IMatchCollection_Release(matches);
                return hres;
            }

            hres = IHTMLTxtRange_findText(range, match_str,
                                          ctx->dir, ctx->flags, &find_res);
            SysFreeString(match_str);
            if(FAILED(hres)) {
                IMatchCollection_Release(matches);
                return hres;
            }

            if(find_res != VARIANT_FALSE) {
                *res = VARIANT_TRUE;
                break;
            }
        }
        if(i == cnt) {
            *res = VARIANT_FALSE;
        }
    }

    IMatchCollection_Release(matches);

    return S_OK;
}


static void set_status_text(wnd_ctx_t *ctx)
{
    TCHAR buf[256];

    wsprintf(buf, _T("%s%sI-search%s:"),
             ctx->failing ? "failing " : "",
             ctx->wrapped ? "wrapped " : "",
             ctx->dir > 0 ? "" : " backward");
    buf[0] = _totupper(buf[0]);
    SetWindowText(ctx->stat, buf);
}

static void search_word(wnd_ctx_t *ctx, int offset)
{
    HRESULT hres;
    IHTMLTxtRange *search_range = NULL;

    if(wcslen(ctx->word) == 0) {
        return;
    }

    {
        IHTMLElement *elem;
        IHTMLBodyElement *body;

        hres = IHTMLDocument2_get_body(ctx->doc, &elem);
        if(FAILED(hres)) {
            error_message("IHTMLDocument2_get_body failed", hres);
            return;
        }

        hres = IHTMLElement_QueryInterface(elem, &IID_IHTMLBodyElement,
                                           (void **)(void *)&body);
        IHTMLElement_Release(elem);
        if(FAILED(hres)) {
            error_message("IHTMLElement_QueryInterface failed", hres);
            return;
        }

        hres = IHTMLBodyElement_createTextRange(body, &search_range);
        IHTMLBodyElement_Release(body);
        if(FAILED(hres)) {
            error_message("IHTMLBodyElement_createTextRange failed", hres);
            return;
        }
    }

    if(ctx->failing) {
        ctx->failing = FALSE;
    } else {
        IHTMLTxtRange *sel_range;

        hres = get_selected_range(ctx->doc, &sel_range);
        if(SUCCEEDED(hres)) {
            BSTR str;
            long cnt;

            str = SysAllocString(ctx->dir > 0 ?
                                 OLESTR("StartToStart") : OLESTR("EndToEnd"));
            hres = IHTMLTxtRange_setEndPoint(search_range, str, sel_range);
            SysFreeString(str);
            IHTMLTxtRange_Release(sel_range);
            if(FAILED(hres)) {
                error_message("IHTMLTxtRange_setEndPoint failed", hres);
                goto func_end;
            }

            str = SysAllocString(OLESTR("character"));
            if(ctx->dir > 0) {
                hres = IHTMLTxtRange_moveStart(search_range, str,
                                               offset, &cnt);
            } else {
                hres = IHTMLTxtRange_moveEnd(search_range, str,
                                             offset, &cnt);
            }
            SysFreeString(str);
            if(FAILED(hres)) {
                error_message("IHTMLTxtRange_move(Start|End) failed", hres);
                goto func_end;
            }
        }
    }

    {
        VARIANT_BOOL res;

        hres = find_text_with_migemo(ctx, search_range,
                                     &res);
        if(FAILED(hres)) {
            error_message("find_text_with_migemo failed", hres);
            goto func_end;
        }

        if(res == VARIANT_FALSE) {
            ctx->failing = TRUE;
        } else {
            hres = IHTMLTxtRange_select(search_range);
            if(FAILED(hres)) {
                error_message("IHTMLTxtRange_select failed", hres);
                goto func_end;
            }
        }
    }

    set_status_text(ctx);

  func_end:

    if(search_range != NULL) IHTMLTxtRange_Release(search_range);
}


static LRESULT create_proc(HWND hwnd, UINT msg,
                           WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    BSTR word;
    int w, h;

    ctx->main = hwnd;

    {
        RECT rect;

        GetClientRect(ctx->main, &rect);
        w = rect.right - margin * 2;
        h = rect.bottom - margin * 2;
    }

    ctx->stat = CreateWindow(_T("Static"), _T(""),
                             WS_VISIBLE | WS_CHILD,
                             margin, margin, (int)(w * 0.4), h,
                             ctx->main, NULL, NULL, NULL);
    if(ctx->stat == NULL) {
        return -1;
    }

    ctx->edit = CreateWindowW(L"Edit", L"",
                              WS_VISIBLE | WS_CHILD,
                              margin + (int)(w * 0.4), margin,
                              (int)(w * 0.6), h,
                              ctx->main, NULL, NULL, NULL);
    if(ctx->edit == NULL) {
        return -1;
    }

    if(FAILED(get_selected_text(ctx->doc, &word)) || word == NULL) {
        word = SysAllocString(OLESTR(""));
        if(word == NULL) {
            return -1;
        }
    }
    ctx->word = word;

    SendMessage(ctx->stat, WM_SETFONT,
                (WPARAM)GetStockObject(DEFAULT_GUI_FONT), 0);
    set_status_text(ctx);

    SendMessageW(ctx->edit, WM_SETFONT,
                 (WPARAM)GetStockObject(DEFAULT_GUI_FONT), 0);
    SetWindowTextW(ctx->edit, ctx->word);
    SetFocus(ctx->edit);
    if(wcslen(ctx->word) != 0) {
        SendMessageW(ctx->edit, EM_SETSEL, 0, -1);
    }

    return 0;
}

static LRESULT destroy_proc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    SysFreeString(ctx->word);

    return 0;
}

static LRESULT close_proc(HWND hwnd, UINT msg,
                          WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    DestroyWindow(hwnd);
    ctx->main = NULL;

    return 0;
}

static LRESULT edit_change(HWND hwnd, UINT msg,
                           WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    BSTR word;
    int len;

    len = GetWindowTextLengthW(ctx->edit);
    word = SysAllocStringLen(NULL, len);
    if(word == NULL) {
        return 0;
    }

    GetWindowTextW(ctx->edit, word, len + 1);

    SysFreeString(ctx->word);
    ctx->word = word;

    ctx->failing = FALSE;
    if(ctx->dir > 0) {
        search_word(ctx, 0);
    } else {
        search_word(ctx, +1);
    }

    return 0;
}

static LRESULT edit_killfocus(HWND hwnd, UINT msg,
                              WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    PostMessage(ctx->main, WM_APP_CLOSE, 0, 0);

    return 0;
}

static LRESULT command_proc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    WORD code;
    msg_proc_t proc;

    static struct uint_ptr_pair cmd_proc[] = {
        {EN_CHANGE, edit_change},
        {EN_KILLFOCUS, edit_killfocus},

        {0, NULL}
    };

    code = HIWORD(wparam);

    proc = (msg_proc_t)assq_pair(cmd_proc, code, NULL);
    if(proc == NULL) {
        return 0;
    }

    return proc(hwnd, msg, wparam, lparam, ctx);
}

static LRESULT forward_proc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    if(ctx->dir > 0 && ctx->failing) {
        ctx->wrapped = TRUE;
    }
    ctx->dir = +1;
    search_word(ctx, +1);

    return 0;
}

static LRESULT backward_proc(HWND hwnd, UINT msg,
                             WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    if(ctx->dir < 0 && ctx->failing) {
        ctx->wrapped = TRUE;
    }
    ctx->dir = -1;
    search_word(ctx, -1);

    return 0;
}

static LRESULT sel_click(HWND hwnd, UINT msg,
                         WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    IHTMLTxtRange *sel_range;
    IHTMLElement *sel_elem;
    HRESULT hres;

    hres = get_selected_range(ctx->doc, &sel_range);
    if(FAILED(hres)) {
        error_message("get_selected_range failed", hres);
        return 0;
    }

    hres = IHTMLTxtRange_parentElement(sel_range, &sel_elem);
    IHTMLTxtRange_Release(sel_range);
    if(FAILED(hres)) {
        error_message("IHTMLTxtRange_parentElement failed", hres);
        return 0;
    }

    hres = IHTMLElement_click(sel_elem);
    IHTMLElement_Release(sel_elem);
    if(FAILED(hres)) {
        error_message("IHTMLElement_click failed", hres);
        return 0;
    }

    return 0;
}

static LRESULT default_proc(HWND hwnd, UINT msg,
                            WPARAM wparam, LPARAM lparam, wnd_ctx_t *ctx)
{
    return DefWindowProc(hwnd, msg, wparam, lparam);
}

static LRESULT CALLBACK window_proc(HWND hwnd, UINT msg,
                                    WPARAM wparam, LPARAM lparam)
{
    msg_proc_t proc;
    wnd_ctx_t *ctx;

    static struct uint_ptr_pair msg_proc[] = {
        {WM_CREATE, create_proc},
        {WM_DESTROY, destroy_proc},
        {WM_APP_CLOSE, close_proc},
        {WM_COMMAND, command_proc},
        {WM_APP_FWD, forward_proc},
        {WM_APP_BWD, backward_proc},
        {WM_APP_SEL_CLICK, sel_click},

        {0, NULL}
    };

    if(msg == WM_CREATE) {
        CREATESTRUCT *cs;

        cs = (CREATESTRUCT *)lparam;
        ctx = (wnd_ctx_t *)cs->lpCreateParams;
        SetWindowLongPtr(hwnd, 0, (LONG_PTR)ctx);
    } else {
        ctx = (wnd_ctx_t *)GetWindowLongPtr(hwnd, 0);
    }

    proc = (msg_proc_t)assq_pair(msg_proc, msg, default_proc);

    return proc(hwnd, msg, wparam, lparam, ctx);
}


static HWND create_window(wnd_ctx_t *param, HWND parent)
{
    HWND hwnd;
    int ret;

    {
        WNDCLASSEX wc;

        memset(&wc, 0, sizeof(wc));
        wc.cbSize = sizeof(wc);
        wc.style = 0;
        wc.lpfnWndProc = window_proc;
        wc.cbClsExtra = 0;
        wc.cbWndExtra = sizeof(wnd_ctx_t *);
        wc.hInstance = instance;
        wc.hIcon = NULL;
        wc.hCursor = NULL;
        wc.hbrBackground = (HBRUSH)(COLOR_BTNFACE + 1);
        wc.lpszMenuName = NULL;
        wc.lpszClassName = window_class_name;
        wc.hIconSm = NULL;

        ret = RegisterClassEx(&wc);
        if(ret == 0) {
            return NULL;
        }
    }

    {
        RECT rect;
        int x, y, w, h;
        LOGFONT logfont;

        GetObject(GetStockObject(DEFAULT_GUI_FONT), sizeof(logfont), &logfont);
        h = logfont.lfHeight;
        if(h < 0) {
            HDC dc;

            dc = GetDC(NULL);
            h = -MulDiv(h, GetDeviceCaps(dc, LOGPIXELSY), 72);
            ReleaseDC(NULL, dc);
        }

        h = h + margin * 2;

        GetWindowRect(parent, &rect);
        w = rect.right - rect.left;
        x = rect.left;
        y = rect.bottom - h;

        hwnd = CreateWindow(window_class_name, window_title_name,
                            WS_POPUP,
                            x, y, w, h,
                            parent, NULL, instance, param);
    }

    return hwnd;
}

static void message_loop(wnd_ctx_t *ctx)
{
    MSG msg;
    int ret;

    while(ctx->main != NULL) {
        ret = GetMessage(&msg, NULL, 0, 0);
        if(ret < 0) {
            break;
        }

        if(ret == 0) {
            break;
        }

        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
}


void isearch_html_doc(IHTMLDocument2 *doc, HWND parent,
                      int dir, int ignore_case,
                      char *migemo_dict)
{
    IHTMLDocument2 *active_doc;
    wnd_ctx_t ctx;
    HWND hwnd = NULL;
    HRESULT hres;

    hres = get_active_document(doc, &active_doc);
    if(FAILED(hres)) {
        error_message("get_active_document failed", hres);
        return;
    }

    memset(&ctx, 0, sizeof(ctx));
    ctx.doc = active_doc;
    ctx.dir = dir;
    ctx.flags = 0;

    {
        hres = CoCreateInstance(&CLSID_RegExp, NULL, CLSCTX_INPROC_SERVER,
                                &IID_IRegExp, (void **)&ctx.regexp);
        if(FAILED(hres)) {
            error_message("CoCreateInstance failed", hres);
            goto func_end;
        }

        hres = IRegExp_put_Global(ctx.regexp, VARIANT_TRUE);
        if(FAILED(hres)) {
            error_message("IRegExp_put_Global failed", hres);
            goto func_end;
        }

        hres = IRegExp_put_IgnoreCase(ctx.regexp,
                                      (ignore_case ?
                                       VARIANT_TRUE : VARIANT_FALSE));
        if(FAILED(hres)) {
            error_message("IRegExp_put_IgnoreCase failed", hres);
            goto func_end;
        }
    }

    {
        hwnd = create_window(&ctx, parent);
        if(hwnd == NULL) {
            error_message("create_window failed", GetLastError());
            goto func_end;
        }

        ShowWindow(hwnd, SW_SHOW);
    }

    {
        ctx.mp = migemo_open(migemo_dict);
        if(ctx.mp == NULL) {
            error_message("migemo_open failed", errno);
            goto func_end;
        }
    }

    message_loop(&ctx);

  func_end:
    if(ctx.main != NULL) DestroyWindow(ctx.main);
    if(ctx.regexp != NULL) IRegExp_Release(ctx.regexp);
    if(ctx.mp != NULL) migemo_close(ctx.mp);
    UnregisterClass(window_class_name, instance);
    IHTMLDocument2_Release(active_doc);
}
